@@ -1,16 +1,17 @@
 import ast
-from fnmatch import fnmatch, fnmatchcase
-from pathlib import Path
 import pkgutil
 import re
 import string
 import sys
+from fnmatch import fnmatch, fnmatchcase
+from functools import partial
+from pathlib import Path
+from typing import List
 
-from vulture import lines
-from vulture import noqa
-from vulture import utils
-from vulture.config import make_config
-
+from vulture import lines, noqa, utils
+from vulture.config import InputError, make_config
+from vulture.reachability import Reachability
+from vulture.utils import ExitCode
 
 DEFAULT_CONFIDENCE = 60
 
@@ -138,13 +139,13 @@ class Item:
         message="",
         confidence=DEFAULT_CONFIDENCE,
     ):
-        self.name = name
-        self.typ = typ
-        self.filename = filename
-        self.first_lineno = first_lineno
-        self.last_lineno = last_lineno
-        self.message = message or f"unused {typ} '{name}'"
-        self.confidence = confidence
+        self.name: str = name
+        self.typ: str = typ
+        self.filename: Path = filename
+        self.first_lineno: int = first_lineno
+        self.last_lineno: int = last_lineno
+        self.message: str = message or f"unused {typ} '{name}'"
+        self.confidence: int = confidence
 
     @property
     def size(self):
@@ -157,12 +158,9 @@ class Item:
             size_report = f", {self.size:d} {line_format}"
         else:
             size_report = ""
-        return "{}:{:d}: {} ({}% confidence{})".format(
-            utils.format_path(self.filename),
-            self.first_lineno,
-            self.message,
-            self.confidence,
-            size_report,
+        return (
+            f"{utils.format_path(self.filename)}:{self.first_lineno:d}: "
+            f"{self.message} ({self.confidence}% confidence{size_report})"
         )
 
     def get_whitelist_string(self):
@@ -173,8 +171,9 @@ class Item:
             prefix = ""
             if self.typ in ["attribute", "method", "property"]:
                 prefix = "_."
-            return "{}{}  # unused {} ({}:{:d})".format(
-                prefix, self.name, self.typ, filename, self.first_lineno
+            return (
+                f"{prefix}{self.name}  # unused {self.typ} "
+                f"({filename}:{self.first_lineno:d})"
             )
 
     def _tuple(self):
@@ -217,7 +216,15 @@ class Vulture(ast.NodeVisitor):
 
         self.filename = Path()
         self.code = []
-        self.found_dead_code_or_error = False
+        self.exit_code = ExitCode.NoDeadCode
+        self.noqa_lines = {}
+
+        report = partial(
+            self._define,
+            collection=self.unreachable_code,
+            confidence=100,
+        )
+        self.reachability = Reachability(report=report)
 
     def scan(self, code, filename=""):
         filename = Path(filename)
@@ -227,29 +234,27 @@ class Vulture(ast.NodeVisitor):
 
         def handle_syntax_error(e):
             text = f' at "{e.text.strip()}"' if e.text else ""
-            print(
+            self._log(
                 f"{utils.format_path(filename)}:{e.lineno}: {e.msg}{text}",
                 file=sys.stderr,
+                force=True,
             )
-            self.found_dead_code_or_error = True
+            self.exit_code = ExitCode.InvalidInput
 
         try:
-            node = (
-                ast.parse(
-                    code, filename=str(self.filename), type_comments=True
-                )
-                if sys.version_info >= (3, 8)  # type_comments requires 3.8+
-                else ast.parse(code, filename=str(self.filename))
+            node = ast.parse(
+                code, filename=str(self.filename), type_comments=True
             )
         except SyntaxError as err:
             handle_syntax_error(err)
         except ValueError as err:
             # ValueError is raised if source contains null bytes.
-            print(
+            self._log(
                 f'{utils.format_path(filename)}: invalid source code "{err}"',
                 file=sys.stderr,
+                force=True,
             )
-            self.found_dead_code_or_error = True
+            self.exit_code = ExitCode.InvalidInput
         else:
             # When parsing type comments, visiting can throw SyntaxError.
             try:
@@ -257,6 +262,10 @@ class Vulture(ast.NodeVisitor):
             except SyntaxError as err:
                 handle_syntax_error(err)
 
+        # Reset the reachability internals for every module to reduce memory
+        # usage.
+        self.reachability.reset()
+
     def scavenge(self, paths, exclude=None):
         def prepare_pattern(pattern):
             if not any(char in pattern for char in "*?["):
@@ -279,12 +288,13 @@ class Vulture(ast.NodeVisitor):
             try:
                 module_string = utils.read_file(module)
             except utils.VultureInputException as err:  # noqa: F841
-                print(
+                self._log(
                     f"Error: Could not read file {module} - {err}\n"
                     f"Try to change the encoding to UTF-8.",
                     file=sys.stderr,
+                    force=True,
                 )
-                self.found_dead_code_or_error = True
+                self.exit_code = ExitCode.InvalidInput
             else:
                 self.scan(module_string, filename=module)
 
@@ -300,10 +310,13 @@ class Vulture(ast.NodeVisitor):
                 except OSError:
                     # Most imported modules don't have a whitelist.
                     continue
+                assert module_data is not None
                 module_string = module_data.decode("utf-8")
                 self.scan(module_string, filename=path)
 
-    def get_unused_code(self, min_confidence=0, sort_by_size=False):
+    def get_unused_code(
+        self, min_confidence=0, sort_by_size=False
+    ) -> List[Item]:
         """
         Return ordered list of unused Item objects.
         """
@@ -344,13 +357,14 @@ class Vulture(ast.NodeVisitor):
         for item in self.get_unused_code(
             min_confidence=min_confidence, sort_by_size=sort_by_size
         ):
-            print(
+            self._log(
                 item.get_whitelist_string()
                 if make_whitelist
-                else item.get_report(add_size=sort_by_size)
+                else item.get_report(add_size=sort_by_size),
+                force=True,
             )
-            self.found_dead_code_or_error = True
-        return self.found_dead_code_or_error
+            self.exit_code = ExitCode.DeadCode
+        return self.exit_code
 
     @property
     def unused_classes(self):
@@ -380,9 +394,15 @@ class Vulture(ast.NodeVisitor):
     def unused_attrs(self):
         return _get_unused_items(self.defined_attrs, self.used_names)
 
-    def _log(self, *args):
-        if self.verbose:
-            print(*args)
+    def _log(self, *args, file=None, force=False):
+        if self.verbose or force:
+            file = file or sys.stdout
+            try:
+                print(*args, file=file)
+            except UnicodeEncodeError:
+                # Some terminals can't print Unicode symbols.
+                x = " ".join(map(str, args))
+                print(x.encode(), file=file)
 
     def _add_aliases(self, node):
         """
@@ -405,47 +425,6 @@ class Vulture(ast.NodeVisitor):
             if alias is not None:
                 self.used_names.add(name_and_alias.name)
 
-    def _handle_conditional_node(self, node, name):
-        if utils.condition_is_always_false(node.test):
-            self._define(
-                self.unreachable_code,
-                name,
-                node,
-                last_node=node.body
-                if isinstance(node, ast.IfExp)
-                else node.body[-1],
-                message=f"unsatisfiable '{name}' condition",
-                confidence=100,
-            )
-        elif utils.condition_is_always_true(node.test):
-            else_body = node.orelse
-            if name == "ternary":
-                self._define(
-                    self.unreachable_code,
-                    name,
-                    else_body,
-                    message="unreachable 'else' expression",
-                    confidence=100,
-                )
-            elif else_body:
-                self._define(
-                    self.unreachable_code,
-                    "else",
-                    else_body[0],
-                    last_node=else_body[-1],
-                    message="unreachable 'else' block",
-                    confidence=100,
-                )
-            elif name == "if":
-                # Redundant if-condition without else block.
-                self._define(
-                    self.unreachable_code,
-                    name,
-                    node,
-                    message="redundant if-condition",
-                    confidence=100,
-                )
-
     def _define(
         self,
         collection,
@@ -470,14 +449,13 @@ class Vulture(ast.NodeVisitor):
         if ignored(first_lineno):
             self._log(f'Ignoring {typ} "{name}"')
         else:
-            last_lineno = lines.get_last_line_number(last_node)
             collection.append(
                 Item(
                     name,
                     typ,
                     self.filename,
                     first_lineno,
-                    last_lineno,
+                    lines.get_last_line_number(last_node),
                     message=message,
                     confidence=confidence,
                 )
@@ -512,11 +490,11 @@ class Vulture(ast.NodeVisitor):
         "%(my_var)s" % locals()
         """
         if (
-            isinstance(node.left, ast.Str)
+            utils.is_ast_string(node.left)
             and isinstance(node.op, ast.Mod)
             and self._is_locals_call(node.right)
         ):
-            self.used_names |= set(re.findall(r"%\((\w+)\)", node.left.s))
+            self.used_names |= set(re.findall(r"%\((\w+)\)", node.left.value))
 
     def visit_Call(self, node):
         # Count getattr/hasattr(x, "some_attr", ...) as usage of some_attr.
@@ -525,21 +503,21 @@ class Vulture(ast.NodeVisitor):
             or (node.func.id == "hasattr" and len(node.args) == 2)
         ):
             attr_name_arg = node.args[1]
-            if isinstance(attr_name_arg, ast.Str):
-                self.used_names.add(attr_name_arg.s)
+            if utils.is_ast_string(attr_name_arg):
+                self.used_names.add(attr_name_arg.value)
 
         # Parse variable names in new format strings:
         # "{my_var}".format(**locals())
         if (
             isinstance(node.func, ast.Attribute)
-            and isinstance(node.func.value, ast.Str)
+            and utils.is_ast_string(node.func.value)
             and node.func.attr == "format"
             and any(
                 kw.arg is None and self._is_locals_call(kw.value)
                 for kw in node.keywords
             )
         ):
-            self._handle_new_format_string(node.func.value.s)
+            self._handle_new_format_string(node.func.value.value)
 
     def _handle_new_format_string(self, s):
         def is_identifier(name):
@@ -619,12 +597,6 @@ class Vulture(ast.NodeVisitor):
                 self.defined_funcs, node.name, node, ignore=_ignore_function
             )
 
-    def visit_If(self, node):
-        self._handle_conditional_node(node, "if")
-
-    def visit_IfExp(self, node):
-        self._handle_conditional_node(node, "ternary")
-
     def visit_Import(self, node):
         self._add_aliases(node)
 
@@ -645,17 +617,19 @@ class Vulture(ast.NodeVisitor):
         if _assigns_special_variable__all__(node):
             assert isinstance(node.value, (ast.List, ast.Tuple))
             for elt in node.value.elts:
-                if isinstance(elt, ast.Str):
-                    self.used_names.add(elt.s)
-
-    def visit_While(self, node):
-        self._handle_conditional_node(node, "while")
+                if utils.is_ast_string(elt):
+                    self.used_names.add(elt.value)
 
     def visit_MatchClass(self, node):
         for kwd_attr in node.kwd_attrs:
             self.used_names.add(kwd_attr)
 
     def visit(self, node):
+        # Visit children nodes first to allow recursive reachability analysis.
+        self.generic_visit(node)
+
+        self.reachability.visit(node)
+
         method = "visit_" + node.__class__.__name__
         visitor = getattr(self, method, None)
         if self.verbose:
@@ -678,36 +652,10 @@ class Vulture(ast.NodeVisitor):
                 ast.parse(type_comment, filename="<type_comment>", mode=mode)
             )
 
-        return self.generic_visit(node)
-
-    def _handle_ast_list(self, ast_list):
-        """
-        Find unreachable nodes in the given sequence of ast nodes.
-        """
-        for index, node in enumerate(ast_list):
-            if isinstance(
-                node, (ast.Break, ast.Continue, ast.Raise, ast.Return)
-            ):
-                try:
-                    first_unreachable_node = ast_list[index + 1]
-                except IndexError:
-                    continue
-                class_name = node.__class__.__name__.lower()
-                self._define(
-                    self.unreachable_code,
-                    class_name,
-                    first_unreachable_node,
-                    last_node=ast_list[-1],
-                    message=f"unreachable code after '{class_name}'",
-                    confidence=100,
-                )
-                return
-
     def generic_visit(self, node):
         """Called if no explicit visitor function exists for a node."""
         for _, value in ast.iter_fields(node):
             if isinstance(value, list):
-                self._handle_ast_list(value)
                 for item in value:
                     if isinstance(item, ast.AST):
                         self.visit(item)
@@ -716,7 +664,12 @@ class Vulture(ast.NodeVisitor):
 
 
 def main():
-    config = make_config()
+    try:
+        config = make_config()
+    except InputError as e:
+        print(e, file=sys.stderr)
+        sys.exit(ExitCode.InvalidCmdlineArguments)
+
     vulture = Vulture(
         verbose=config["verbose"],
         ignore_names=config["ignore_names"],
