@@ -2,12 +2,20 @@ import ast
 import pathlib
 import sys
 import tokenize
+from enum import IntEnum
 
 
 class VultureInputException(Exception):
     pass
 
 
+class ExitCode(IntEnum):
+    NoDeadCode = 0
+    InvalidInput = 1
+    InvalidCmdlineArguments = 2
+    DeadCode = 3
+
+
 def _safe_eval(node, default):
     """
     Safely evaluate the Boolean expression under the given AST node.
@@ -45,6 +53,10 @@ def condition_is_always_true(condition):
     return _safe_eval(condition, False)
 
 
+def is_ast_string(node):
+    return isinstance(node, ast.Constant) and isinstance(node.value, str)
+
+
 def format_path(path):
     try:
         return path.relative_to(pathlib.Path.cwd())
@@ -56,20 +68,23 @@ def format_path(path):
 def get_decorator_name(decorator):
     if isinstance(decorator, ast.Call):
         decorator = decorator.func
-    parts = []
-    while isinstance(decorator, ast.Attribute):
-        parts.append(decorator.attr)
-        decorator = decorator.value
-    parts.append(decorator.id)
+    try:
+        parts = []
+        while isinstance(decorator, ast.Attribute):
+            parts.append(decorator.attr)
+            decorator = decorator.value
+        parts.append(decorator.id)
+    except AttributeError:
+        parts = []
     return "@" + ".".join(reversed(parts))
 
 
 def get_modules(paths):
     """Retrieve Python files to check.
 
-    Loop over all given paths, abort if any ends with .pyc and add collect
-    the other given files (even those not ending with .py) and all .py
-    files under the given directories.
+    Loop over all given paths, abort if any ends with .pyc, add the other given
+    files (even those not ending with .py) and collect all .py files under the
+    given directories.
 
     """
     modules = []
@@ -93,7 +108,7 @@ def read_file(filename):
         with tokenize.open(filename) as f:
             return f.read()
     except (SyntaxError, UnicodeDecodeError) as err:
-        raise VultureInputException(err)
+        raise VultureInputException from err
 
 
 class LoggingList(list):
